import pytest
import os
import logging
import json
import sys
from time import sleep
from datetime import datetime

sys.path.insert(0, 'producer/api/python/')
sys.path.insert(0, 'consumer/api/python/')

import asapo_producer
import asapo_consumer

log = logging.getLogger(__name__)


def callback(payload, err):
    if err is not None and not isinstance(err, asapo_producer.AsapoServerWarning):
        # the data was not sent. Something is terribly wrong.
        log.debug(f"could not send: {payload} {err}")
    elif err is not None:
        # The data was sent, but there was some unexpected problem, e.g. the file was overwritten.
        log.debug(f"sent with warning: {payload} {err}")
    else:
        # all fine
        log.debug(f"successfuly sent: {payload}")


@pytest.fixture(autouse=True)
def cleanup_db(asapo_config):
    log.info(f"Run test for consumer {asapo_consumer.__version__} and producer {asapo_producer.__version__}")
    os.makedirs(asapo_config.data_path, exist_ok=True)
    yield None
    os.system(f'echo "db.dropDatabase()" | mongo {asapo_config.beamtime}_{asapo_config.data_source} >/dev/null')
    os.system(f"rm -fr {asapo_config.data_path}")


@pytest.mark.compatible
def test_delete_stream(asapo_config):
    stream_name = "test001"
    ts = datetime.now().strftime('%Y%m%dT_%H%M%S')
    data_source = f"source_{ts}"
    producer = asapo_producer.create_producer(asapo_config.host, 'processed', asapo_config.beamtime, "auto",
                                              data_source,
                                              asapo_config.token, 1, 6000)

    consumer = asapo_consumer.create_consumer(asapo_config.host, "", False, asapo_config.beamtime,
                                              data_source, asapo_config.token, 1000)
    g = consumer.generate_group_id()

    def inject_data():
        for i in range(5):
            producer.send(i + 1,  # message number. Should be unique and ordered.
                          f"processed/{stream_name}_file_part{i}.cbf",  #
                          b"The data",  # binary data
                          callback=callback,
                          stream=stream_name)
        producer.wait_requests_finished(3000)

    # Create stream by injecting data
    inject_data()
    # asapo_consumer.AsapoEndOfStreamError
    # check stream info
    log.debug(consumer.get_stream_list())
    stream = consumer.get_stream_list()[0]
    assert (stream['name'] == stream_name)
    assert (stream['lastId'] == 5)
    log.debug(stream)

    # Delete stream
    producer.delete_stream(stream=stream_name)
    # There is no wait or call-back function. Wait manually.
    sleep(1)
    stream_list = consumer.get_stream_list()
    assert (len(stream_list) == 0)

    # Create stream again
    inject_data()
    stream = consumer.get_stream_list()[0]
    assert (stream['name'] == stream_name)
    assert (stream['lastId'] == 5)
    log.debug(stream)

    # Delete stream with consumer
    consumer.delete_stream(stream=stream_name)
    sleep(1)
    stream_list = consumer.get_stream_list()
    assert (len(stream_list) == 0)


@pytest.mark.compatible
def test_get_next_ordered(asapo_config):
    stream_name = "test001"
    ts = datetime.now().strftime('%Y%m%dT_%H%M%S')
    data_source = f"source_{ts}"
    producer = asapo_producer.create_producer(asapo_config.host, 'processed', asapo_config.beamtime, "auto",
                                              data_source,
                                              asapo_config.token, 1, 6000)

    consumer = asapo_consumer.create_consumer(asapo_config.host, "", False, asapo_config.beamtime,
                                              data_source, asapo_config.token, 1000)
    producer.send(2,
                  f"processed/{stream_name}_file_part1.cbf",  #
                  b"The data",  # binary data
                  callback=callback,
                  stream=stream_name)
    producer.wait_requests_finished(3000)

    with pytest.raises(Exception):
        consumer.get_next(consumer.generate_group_id(), meta_only=True, stream=stream_name)

    producer.send(1,
                  f"processed/{stream_name}_file_part1.cbf",  #
                  b"The data",  # binary data
                  callback=callback,
                  stream=stream_name)
    producer.wait_requests_finished(3000)

    _, mdata = consumer.get_next(consumer.generate_group_id(), meta_only=True, stream=stream_name)
    assert (mdata['_id'] == 1)


@pytest.mark.compatible
def test_get_next_ordered_resend_no_delay(asapo_config):
    ts = datetime.now().strftime('%Y%m%dT_%H%M%S')
    stream_name = f"test001_{ts}"
    producer = asapo_producer.create_producer(asapo_config.host, 'processed', asapo_config.beamtime, "auto",
                                              asapo_config.data_source,
                                              asapo_config.token, 1, 6000)

    consumer = asapo_consumer.create_consumer(asapo_config.host, "", False, asapo_config.beamtime,
                                              asapo_config.data_source, asapo_config.token, 500)
    g = consumer.generate_group_id()
    consumer.set_resend_nacs(True, 0, 2)

    producer.send(2,
                  f"processed/{stream_name}_file_part1.cbf",  #
                  b"The data",  # binary data
                  callback=callback,
                  stream=stream_name)
    producer.wait_requests_finished(3000)

    with pytest.raises(asapo_consumer.AsapoNoDataError):
        consumer.get_next(g, meta_only=True, stream=stream_name)

    with pytest.raises(asapo_consumer.AsapoNoDataError):
        consumer.get_next(g, meta_only=True, stream=stream_name)

    producer.send(1,
                  f"processed/{stream_name}_file_part1.cbf",  #
                  b"The data",  # binary data
                  callback=callback,
                  stream=stream_name)
    producer.wait_requests_finished(3000)

    _, mdata = consumer.get_next(g, meta_only=True, stream=stream_name)
    assert (mdata['_id'] == 1)


def test_get_next_available_resend_no_delay(asapo_config):
    ts = datetime.now().strftime('%Y%m%dT_%H%M%S')
    stream_name = f"test001_{ts}"
    producer = asapo_producer.create_producer(asapo_config.host, 'processed', asapo_config.beamtime, "auto",
                                              asapo_config.data_source,
                                              asapo_config.token, 1, 6000)

    consumer = asapo_consumer.create_consumer(asapo_config.host, "", False, asapo_config.beamtime,
                                              asapo_config.data_source, asapo_config.token, 500)
    g = consumer.generate_group_id()
    consumer.set_resend_nacs(True, 0, 2)

    producer.send(2,
                  f"processed/{stream_name}_file_part1.cbf",  #
                  b"The data",  # binary data
                  callback=callback,
                  stream=stream_name)
    producer.wait_requests_finished(3000)

    producer.send(1,
                  f"processed/{stream_name}_file_part1.cbf",  #
                  b"The data",  # binary data
                  callback=callback,
                  stream=stream_name)
    producer.wait_requests_finished(3000)

    _, mdata = consumer.get_next(g, meta_only=True, stream=stream_name, ordered=False)
    assert (mdata['_id'] == 2)


@pytest.mark.compatible
def test_get_by_id_and_last(asapo_config):
    ts = datetime.now().strftime('%Y%m%dT_%H%M%S')
    stream_name = f"test001_{ts}"
    producer = asapo_producer.create_producer(asapo_config.host, 'processed', asapo_config.beamtime, "auto",
                                              asapo_config.data_source,
                                              asapo_config.token, 1, 6000)

    consumer = asapo_consumer.create_consumer(asapo_config.host, "", False, asapo_config.beamtime,
                                              asapo_config.data_source, asapo_config.token, 500)

    for i in [3, 1, 2, 5, 4]:
        producer.send(i,
                      f"processed/{stream_name}_file_part{i}.cbf",  #
                      b"The data",  # binary data
                      callback=callback,
                      stream=stream_name)
    producer.wait_requests_finished(3000)

    _, mdata = consumer.get_last(meta_only=True, stream=stream_name)
    assert mdata["_id"] == 5

    for i in range(1, 6):
        _, mdata = consumer.get_by_id(i, meta_only=True, stream=stream_name)
        assert mdata["_id"] == i


def test_get_source_list(asapo_config):
    def create_source(source_name):
        ts = datetime.now().strftime('%Y%m%dT_%H%M%S')
        stream_name = f"test001_{ts}"
        producer = asapo_producer.create_producer(asapo_config.host, 'processed', asapo_config.beamtime, "auto",
                                                  source_name, asapo_config.token, 1, 6000)

        producer.send(3,
                      f"processed/{stream_name}_file_part3.cbf",  #
                      b"The data",  # binary data
                      callback=callback,
                      stream="stream_1")
        producer.wait_requests_finished(3000)

    for i in range(5):
        create_source(f"source_{i}")

    consumer = asapo_consumer.create_consumer(asapo_config.host, "", False, asapo_config.beamtime,
                                              asapo_config.data_source, asapo_config.token, 500)

    sources = consumer.get_source_list()
    log.debug(f"All sources {sources}")
    assert len(sources) >= 5  # In case of cleanup is not working
    for i in range(5):
        assert f"source_{i}" in sources
